#!/usr/bin/python
# -*- coding:utf-8 -*-

'''Module that implements the base classes for all clients'''
import hmac
import hashlib
import base64
import time
from uuid import uuid1
import urllib.parse
import sys
from typing import List
import requests

try:
    from kraken.exceptions.exceptions import KrakenExceptions
except ModuleNotFoundError:
    print('USING LOCAL MODULE')
    sys.path.append('/Users/benjamin/repositories/Trading/python-kraken-sdk')
    from kraken.exceptions.exceptions import KrakenExceptions

class KrakenErrorHandler():
    '''Class used to raise an Error or return the response'''

    def __init__(self):
        self.__kexceptions = KrakenExceptions()

    def __get_exception(self, msg):
        return self.__kexceptions.get_exception(msg)

    def check(self, data: dict) -> dict:
        '''Check if the error message is a known Kraken error response
            than raise a custom exception or return the data containing the 'error'
        '''
        if len(data.get('error', [])) == 0 and 'result' in data: return data['result']

        exception = self.__get_exception(data['error'])
        if exception: raise exception(data)
        return data

    def check_send_status(self, data: dict) -> dict:
        '''Used for futures REST responses'''
        if 'sendStatus' in data and 'status' in data['sendStatus']:
            exception = self.__get_exception(data['sendStatus']['status'])
            if exception: raise exception(data)
            return data
        return data

    def check_batch_status(self, data: List[dict]) -> dict:
        '''Used for futures REST batch order responses'''
        if 'batchStatus' in data:
            batch_status = data['batchStatus']
            for status in batch_status:
                if 'status' in status:
                    exception = self.__get_exception(status['status'])
                    if exception: raise exception(data)
        return data

class KrakenBaseRestAPI():
    ''' Base class for all Spot clients

        Handles un/signed requests and returns exception handled results

        ====== P A R A M E T E R S ======
        key: str, defualt: ''
            Spot API public key
        secret: str, default: ''
            Spot API secret key
        url: str, default: 'https://api.kraken.com'
            optional url
        sandbox: bool, default: False
            not used so far
    '''

    URL = 'https://api.kraken.com'
    API_V = '/0'

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False):
        if sandbox: raise ValueError('Sandbox not availabel for Kraken Spot trading.')
        if url != '': self.url = url
        else: self.url = self.URL
        
        self.__key = key
        self.__secret = secret
        self.__err_handler = KrakenErrorHandler()
        self.__session = requests.Session()
        self.__session.headers.update({'User-Agent': 'python-kraken-sdk'})

    def _request(self,
        method: str,
        uri: str,
        timeout: int=10,
        auth: bool=True,
        params: dict=None,
        do_json: bool=False,
        return_raw: bool=False
    ) -> dict:
        if params is None: params = {}
        method = method.upper()
        data_json = ''
        if method in ['GET', 'DELETE']:
            if params:
                strl = []
                for key in sorted(params): strl.append(f'{key}={params[key]}')
                data_json += '&'.join(strl)
                uri += f'?{data_json}'.replace(' ', '%20')

        headers = { }
        if auth:
            if not self.__key or self.__key == '' or not self.__secret or self.__secret == '': raise ValueError('Missing credentials.')
            params['nonce'] = str(int(time.time() * 1000))
            headers.update({
                'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
                'API-Key': self.__key,
                'API-Sign': self.get_kraken_signature(f'{self.API_V}{uri}', params)
            })

        url = f'{self.url}{self.API_V}{uri}'
        if method in ['GET', 'DELETE']:
            return self.__check_response_data(
                self.__session.request(method=method, url=url, headers=headers, timeout=timeout),
                return_raw
            )
        if do_json:
            return self.__check_response_data(
                self.__session.request(method=method, url=url, headers=headers, json=params, timeout=timeout),
                return_raw
            )
        return self.__check_response_data(
            self.__session.request(method=method, url=url, headers=headers, data=params, timeout=timeout),
            return_raw
        )

    def get_kraken_signature(self, urlpath: str, data: dict) -> str:
        '''Returns the signed data'''
        return base64.b64encode(
            hmac.new(
                base64.b64decode(self.__secret),
                urlpath.encode() + hashlib.sha256((str(data['nonce']) + urllib.parse.urlencode(data)).encode()).digest(),
                hashlib.sha512
            ).digest()
        ).decode()

    def __check_response_data(self, response_data, return_raw: bool=False) -> dict:
        '''checkes the response, handles the error and returns the data'''
        if response_data.status_code in [ '200', 200 ]:
            if return_raw: return response_data
            try: data = response_data.json()
            except ValueError as exc: raise ValueError(response_data.content) from exc
            else:
                if 'error' in data: return self.__err_handler.check(data)
                return data
        raise Exception(f'{response_data.status_code} - {response_data.text}')

    @property
    def return_unique_id(self) -> str:
        '''Returns a unique id str'''
        return ''.join(str(uuid1()).split('-'))

    def _to_str_list(self, value) -> str:
        '''Converts a list to a comme separated str'''
        if isinstance(value, str): return value
        if isinstance(value, list): return ','.join(value)
        raise ValueError('a must be type of str or list of strings')

class KrakenBaseFuturesAPI():
    ''' Base class for all Futures clients

        Handles un/signed requests and returns exception handled results

        ====== P A R A M E T E R S ======
        key: str, defualt: ''
            Futures API public key
        secret: str, default: ''
            Futures API secret key
        url: str, default: 'https://futures.kraken.com'
            optional url
        sandbox: bool, default: False
            if set to true the url will be 'https://demo-futures.kraken.com'

        ====== N O T E S ======
        If the sandbox environment is chosen, the keys must be generated here:
            https://demo-futures.kraken.com/settings/api
    '''

    URL = 'https://futures.kraken.com'
    SANDBOX_URL = 'https://demo-futures.kraken.com'

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False):

        self.sandbox = sandbox
        if url: self.url = url
        elif self.sandbox: self.url = self.SANDBOX_URL
        else: self.url = self.URL

        self.__key = key
        self.__secret = secret
        self.__nonce = 0

        self.__err_handler = KrakenErrorHandler()
        self.__session = requests.Session()
        self.__session.headers.update({'User-Agent': 'python-kraken-sdk'})

    def _request(self,
        method: str,
        uri: str,
        timeout: int=10,
        auth: bool=True,
        post_params: dict=None,
        query_params: dict=None,
        return_raw: bool=False
    ) -> dict:
        method = method.upper()

        post_string: str = ''
        if post_params is not None:
            strl: List[str] = []
            for key in sorted(post_params): strl.append(f'{key}={post_params[key]}')
            post_string = '&'.join(strl)
        else: post_params = {}

        query_string: str = ''
        if query_params is not None:
            strl: List[str] = []
            for key in sorted(query_params): strl.append(f'{key}={query_params[key]}')
            query_string = '&'.join(strl).replace(' ', '%20')
        else: query_params = {}

        headers = { }
        if auth:
            if not self.__key or self.__key == '' or not self.__secret or self.__secret == '': raise ValueError('Missing credentials')
            self.__nonce = (self.__nonce + 1) % 1
            nonce = str(int(time.time() * 1000)) + str(self.__nonce).zfill(4)
            headers.update({
                'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
                'Nonce': nonce,
                'APIKey': self.__key,
                'Authent': self.get_kraken_futures_signature(uri, query_string + post_string, nonce)
            })

        if method in ['GET', 'DELETE']:
            return self.__check_response_data(
                self.__session.request(
                    method=method,
                    url=f'{self.url}{uri}' if query_string == '' else f'{self.url}{uri}?{query_string}',
                    headers=headers,
                    timeout=timeout
                ),
                return_raw
            )
        if method == 'PUT':
            return self.__check_response_data(
                self.__session.request(
                method=method,
                url=f'{self.url}{uri}',
                params=str.encode(query_string),
                headers=headers,
                timeout=timeout
                ),
                return_raw
            )

        return self.__check_response_data(
            self.__session.request(
                method=method,
                url=f'{self.url}{uri}?{post_string}',
                data=str.encode(post_string),
                headers=headers,
                timeout=timeout
            ), return_raw
        )

    def get_kraken_futures_signature(self, endpoint: str, data: str, nonce: str) -> str:
        '''
            Returns the signed data/message
            reference: https://github.com/CryptoFacilities/REST-v3-Python/blob/ee89b9b324335d5246e2f3da6b52485eb8391d50/cfRestApiV3.py#L295-L296
        '''
        if endpoint.startswith('/derivatives'): endpoint = endpoint[len('/derivatives'):]
        sha256_hash = hashlib.sha256()
        sha256_hash.update((data + nonce + endpoint).encode('utf8'))
        return base64.b64encode(
            hmac.new(
                base64.b64decode(self.__secret),
                sha256_hash.digest(),
                hashlib.sha512
            ).digest()
        )

    def __check_response_data(self, response_data, return_raw: bool=False) -> dict:
        if response_data.status_code in [ '200', 200 ]:
            if return_raw: return response_data
            try: data = response_data.json()
            except ValueError as exc: raise ValueError(response_data.content) from exc
            else:
                if 'error' in data: return self.__err_handler.check(data)
                if 'sendStatus' in data: return self.__err_handler.check_send_status(data)
                if 'batchStatus' in data: return self.__err_handler.check_batch_status(data)
                return data
        else: raise Exception(f'{response_data.status_code} - {response_data.text}')



'''Module that implements the Spot Kraken Websocket client'''
import logging
from typing import List
from kraken.base_api.base_api import KrakenBaseRestAPI

class SpotWsClientCl(KrakenBaseRestAPI):
    '''Class that implements the Spot Kraken Websocket client'''

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False):
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)

        self._pub_conn = None
        self._priv_conn = None

    def get_ws_token(self) -> dict:
        '''https://docs.kraken.com/rest/#tag/Websockets-Authentication'''
        return self._request('POST', '/private/GetWebSocketsToken')

    async def create_order(
        self,
        ordertype: str,
        side: str,
        pair: str,
        price: str=None,
        price2: str=None,
        volume: str=None,
        leverage: int=None,
        oflags: List[str]=None,
        starttm: str=None,
        expiretm: str=None,
        deadline: str=None,
        userref: str=None,
        validate: str=None,
        close_ordertype: str=None,
        close_price: float=None,
        close_price2: float=None,
        timeinforce: str=None
    ) -> None:
        '''https://docs.kra)en.com/websockets/#message-addOrder'''
        if not self._priv_conn:
            logging.warning('Websocket not connected!')
            return
        if not self._priv_conn.isAuth:
            raise ValueError('Cannot create_order on public Websocket Client!')

        payload = {
            'event': 'addOrder',
            'ordertype': str(ordertype),
            'type': str(side),
            'pair': str(pair),
            'price': str(price)
        }
        if price2 is not None: payload['price2'] = str(price2)
        if volume is not None: payload['volume'] = str(volume)
        if oflags is not None:
            if isinstance(oflags, str): payload['oflags'] = oflags
            elif isinstance(oflags, list): payload['oflags'] = self._to_str_list(oflags)
            else: raise ValueError('oflags must be type List[str] or comma delimited list of order flags as str. Available flags: viqc, fcib, fciq, nompp, post')
        if starttm is not None: payload['starttm'] = starttm
        if expiretm is not None: payload['expiretm'] = expiretm
        if deadline is not None: payload['deadline'] = deadline
        if userref is not None: payload['userref'] = userref
        if validate is not None: payload['validate'] = validate
        if leverage is not None: payload['leverage'] = leverage
        if close_ordertype is not None: payload['close[ordertype]'] = close_ordertype
        if close_price is not None: payload['close[price]'] = close_price
        if close_price2 is not None: payload['close[price2]'] = close_price2
        if timeinforce is not None: payload['timeinforce'] = timeinforce

        await self._priv_conn.send_message(msg=payload, private=True)

    async def edit_order(self,
        orderid: str,
        reqid: int=None,
        pair: str=None,
        price: str=None,
        price2: str=None,
        volume: str=None,
        oflags: List[str]=None,
        newuserref: str=None,
        validate: str=None
    ) -> None:
        '''https://docs.kraken.com/websockets/#message-editOrder'''
        if not self._priv_conn:
            logging.warning('Websocket not connected!')
            return
        if not self._priv_conn.isAuth:
            raise ValueError('Cannot edit_order on public Websocke Client!')

        payload = {
            'event': 'editOrder',
            'orderid': orderid
        }
        if reqid is not None: payload['reqid'] = reqid
        if pair is not None: payload['pair'] = pair
        if price is not None: payload['price'] = str(price)
        if price2 is not None: payload['price2'] = str(price2)
        if volume is not None: payload['volume'] = str(volume)
        if oflags is not None: payload['oflags'] = self._to_str_list(oflags)
        if newuserref is not None: payload['newuserref'] = str(newuserref)
        if validate is not None: payload['validate'] = str(validate)

        await self._priv_conn.send_message(msg=payload, private=True)

    async def cancel_order(self, txid) -> None:
        '''https://docs.kraken.com/websockets/#message-cancelOrder'''
        if not self._priv_conn:
            logging.warning('Websocket not connected!')
            return
        if not self._priv_conn.isAuth:
            raise ValueError('Cannot cancel_order on public Websocke Client!')
        await self._priv_conn.send_message(
            msg={ 'event': 'cancelOrder',  'txid': self._to_str_list(txid) },
            private=True
        )

    async def cancel_all_orders(self) -> None:
        '''https://docs.kraken.com/websockets/#message-cancelAll'''
        if not self._priv_conn:
            logging.warning('Websocket not connected!')
            return
        if not self._priv_conn.isAuth:
            raise ValueError('Cannot use cancel_all_orders on public Websocke Client!')
        await self._priv_conn.send_message(
            msg={ 'event': 'cancelAll' },
            private=True,
            reponse=True
        )

    async def cancel_all_orders_after(self, timeout: int) -> None:
        '''https://docs.kraken.com/websockets/#message-cancelAllOrdersAfter'''
        if not self._priv_conn:
            logging.warning('Websocket not connected!')
            return
        if not self._priv_conn.isAuth:
            raise ValueError('Cannot use cancel_all_orders_after on public Websocke Client!')
        await self._priv_conn.send_message(
            msg={ 'event': 'cancelAllOrdersAfter', 'timeout': timeout },
            private=True
        )



'''
Provides the KrakenWsClientCl class to use the Kraken Futures
websockets.
'''

import hashlib
import hmac
import base64
from kraken.base_api.base_api import KrakenBaseFuturesAPI

class FuturesWsClientCl(KrakenBaseFuturesAPI):
    '''This class will be extended in the future if Kraken implements for example
        sending orders via websocket feed for Futures
    '''
    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False):
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)
        self._conn = None
        self._key = key
        self._secret = secret

    def _get_sign_challenge(self, challenge: str) -> str:
        sha256_hash = hashlib.sha256()
        sha256_hash.update(challenge.encode('utf8'))
        return base64.b64encode(
            hmac.new(
                base64.b64decode(self._secret),
                sha256_hash.digest(),
                hashlib.sha512
            ).digest()
        ).decode('utf-8')



'''Module that implements the Kraken Spot market client'''
from kraken.base_api.base_api import KrakenBaseRestAPI

class MarketClient(KrakenBaseRestAPI):
    '''Class that implements the Kraken Spot market client'''

    def get_assets(self, assets=None, aclass: str=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getAssetInfo'''
        params = {}
        if assets is not None: params['asset'] = self._to_str_list(assets)
        if aclass is not None: params['aclass'] = aclass
        return self._request(method='GET', uri='/public/Assets', params=params, auth=False)

    def get_tradable_asset_pair(self, pair: str, info=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradableAssetPairs'''
        params = {}
        params['pair'] = self._to_str_list(pair)#','.join(pair)
        if info is not None: params['info'] = info

        return self._request(method='GET', uri='/public/AssetPairs', params=params, auth=False)

    def get_ticker(self, pair: str=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTickerInformation'''
        params = { }
        if pair is not None: params['pair'] = self._to_str_list(pair)
        return self._request(method='GET', uri='/public/Ticker', params=params, auth=False)

    def get_ohlc(self, pair: str, interval: int=None, since: int=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getOHLCData
        interval in minutes
        '''
        params = { 'pair': pair }
        if interval is not None: params['interval'] = interval
        if since is not None: params['since'] = since
        return self._request(method='GET', uri='/public/OHLC', params=params, auth=False)

    def get_order_book(self, pair: str, count=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getOrderBook'''
        params = { 'pair': pair }
        if count is not None: params['count'] = count
        return self._request(method='GET', uri='/public/Depth', params=params, auth=False)

    def get_recent_trades(self, pair: str, since=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getRecentTrades'''
        params = { 'pair': pair }
        if since is not None: params['since'] = None
        return self._request(method='GET', uri='/public/Trades', params=params, auth=False)

    def get_recend_spreads(self, pair: str, since=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getRecentSpreads'''
        params = { 'pair': pair }
        if since is not None: params['since'] = since
        return self._request(method='GET', uri='/public/Spread', params=params, auth=False)

    def get_system_status(self) -> dict:
        '''Returns the system status of the Kraken API'''
        return self._request(method='GET', uri='/public/SystemStatus', auth=False)



'''Module that implements the Kraken Futures trade client'''
from typing import List
from kraken.base_api.base_api import KrakenBaseFuturesAPI
class TradeClient(KrakenBaseFuturesAPI):
    '''Class that implements the Kraken Futures trade client'''

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False) -> None:
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)

    def get_fills(self, lastFillTime: str=None) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-historical-data-get-your-fills'''
        query_params = {}
        if lastFillTime: query_params['lastFillTime'] = lastFillTime
        return self._request(method='GET', uri='/derivatives/api/v3/fills', query_params=query_params, auth=True)

    def create_batch_order(self, batchorder_list: List[dict]) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-order-management-batch-order-management'''
        batchorder = { 'batchOrder': batchorder_list }
        return self._request(method='POST', uri='/derivatives/api/v3/batchorder', post_params={ 'json': f'{batchorder}' }, auth=True)

    def cancel_all_orders(self, symbol: str=None) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-order-management-cancel-all-orders'''
        params = {}
        if symbol is not None: params['symbol'] = symbol
        return self._request(method='POST', uri='/derivatives/api/v3/cancelallorders', post_params=params, auth=True)

    def dead_mans_switch(self, timeout: int=60) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-order-management-dead-man-39-s-switch'''
        return self._request(method='POST', uri='/derivatives/api/v3/cancelallordersafter', post_params={'timeout': timeout})

    def cancel_order(self, order_id: str='', cliOrdId: str='') -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-order-management-cancel-order'''
        if order_id == '' and cliOrdId == '': raise ValueError('Either order_id or cliOrdId must be set!')

        params = {}
        if order_id != '': params['order_id'] = order_id
        elif cliOrdId != '': params['cliOrdId'] = cliOrdId
        return self._request(method='POST', uri='/derivatives/api/v3/cancelorder', post_params=params, auth=True)

    def edit_order(self, orderId: str=None, cliOrdId: str=None, limitPrice: float=None, size: float=None, stopPrice: float=None) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-order-management-edit-order'''
        if orderId == '' and cliOrdId == '': raise ValueError('Either orderId or cliOrdId must be set!')

        params = {}
        if orderId != '': params['orderId'] = orderId
        elif cliOrdId != '': params['cliOrdId'] = cliOrdId
        if limitPrice is not None: params['limitPrice'] = limitPrice
        if size is not None: params['size'] = size
        if stopPrice is not None: params['stopPrice'] = stopPrice
        return self._request(method='POST', uri='/derivatives/api/v3/editorder', post_params=params, auth=True)

    def get_orders_status(self, orderIds: List[str]=None, cliOrdIds: List[str]=None) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-order-management-get-the-current-status-for-specific-orders'''
        if orderIds is None and cliOrdIds is None: raise ValueError('Either orderIds or cliOrdIds must be specified!')

        params = {}
        if orderIds is not None: params['orderIds'] = orderIds
        elif cliOrdIds is not None: params['cliOrdIds'] = cliOrdIds
        return self._request(method='POST', uri='/derivatives/api/v3/orders/status', post_params=params, auth=True)

    def create_order(self,
        orderType: str,
        size: float,
        symbol: str,
        side: str,
        cliOrdId: str=None,
        limitPrice: float=None,
        reduceOnly: bool=None,
        stopPrice: float=None,
        triggerSignal: str=None,
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-order-management-send-order'''

        order_types = ['lmt', 'post', 'ioc', 'mkt', 'stp', 'take_profit']
        if orderType not in order_types: raise ValueError(f'Invalid orderType. One of [{order_types}] is required!')
        sides = ['buy', 'sell']
        if side not in sides: raise ValueError(f'Invalid side. One of [{sides}] is required!')

        params = {
            'orderType': orderType,
            'side': side,
            'size': size,
            'symbol': symbol
        }
        if cliOrdId is not None: params['cliOrdId'] = cliOrdId
        if orderType in ['post', 'lmt']:
            if limitPrice is None:
                raise ValueError(f'No limitPrice specified for order of type {orderType}!')
            params['limitPrice'] = limitPrice
        if reduceOnly is not None: params['reduceOnly'] = reduceOnly
        if orderType in ['stp', 'take_profit']:
            if stopPrice is None:
                raise ValueError(f'Parammeter stopPrice must be set if orderType {orderType}!')
            if triggerSignal is None:
                raise ValueError(f'Parammeter triggerSignal must be set if orderType {orderType}!')
        if stopPrice is not None: params['stopPrice'] = stopPrice
        if triggerSignal is not None:
            trigger_signals = ['mark', 'spot', 'last']
            if triggerSignal not in trigger_signals:
                raise ValueError(f'Trigger signal must be in [{trigger_signals}]!')
            params['triggerSignal'] = triggerSignal

        return self._request(method='POST', uri='/derivatives/api/v3/sendorder', post_params=params, auth=True)



'''Module that implements the Spot Funding client'''
from kraken.base_api.base_api import KrakenBaseRestAPI

class FundingClient(KrakenBaseRestAPI):
    '''Class that implements the Spot Funding client'''

    def get_deposit_methods(self, asset: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/getDepositMethods'''
        return self._request(method='POST', uri='/private/DepositMethods', params={ 'asset': asset })

    def get_deposit_address(self, asset: str, method: str, new: bool=False) -> dict:
        '''https://docs.kraken.com/rest/#operation/getDepositAddresses'''
        return self._request(method='POST', uri='/private/DepositAddresses', params={
            'asset': asset,
            'method': method,
            'new': new
        })

    def get_recend_deposits_status(self, asset: str, method: str=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getStatusRecentDeposits'''
        params = { 'asset': asset }
        if method is not None: params['method'] = method
        return self._request(method='POST', uri='/private/DepositStatus', params=params)

    def withdraw_funds(self, asset: str, key: str, amount: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/withdrawFund'''
        return self._request(method='POST', uri='/private/Withdraw', params={
            'asset': asset,
            'key': str(key),
            'amount': str(amount)
        })

    def get_withdrawal_info(self, asset: str, key: str, amount: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/getWithdrawalInformation'''
        return self._request(method='POST', uri='/private/WithdrawInfo', params={
            'asset': asset,
            'key': str(key),
            'amount': str(amount)
        })

    def get_recend_withdraw_status(self, asset: str, method: str=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getStatusRecentWithdrawals'''
        params = { 'asset': asset }
        if method is not None: params['method'] = method
        return self._request(method='POST', uri='/private/WithdrawStatus', params=params)

    def cancel_withdraw(self, asset: str, refid: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/cancelWithdrawal'''
        return self._request(method='POST', uri='/private/WithdrawCancel', params={
            'asset': asset,
            'refid': str(refid)
        })

    def wallet_transfer(self, asset: str, from_: str, to_: str, amount: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/walletTransfer'''
        return self._request(method='POST', uri='/private/WalletTransfer', params={
            'asset': asset,
            'from': from_,
            'to': to_,
            'amount': amount
        })



'''This module provides the Kraken Futures clients'''

from kraken.futures.market.market import MarketClient
from kraken.futures.trade.trade import TradeClient
from kraken.futures.user.user import UserClient
from kraken.futures.funding.funding import FundingClient
from kraken.futures.websocket.websocket import KrakenFuturesWSClientCl


class User(UserClient):
    ''' Class that is used as a client to access the user-related
        Kraken Futures endpoints.
    '''

class Trade(TradeClient):
    ''' Class that is used as a client to access the trade-related
        Kraken Futures endpoints.
    '''

class Market(MarketClient):
    ''' Class that is used as a client to access the market-related
        Kraken Futures endpoints.
    '''

class Funding(FundingClient):
    ''' Class that is used as a client to access the funding-related
        Kraken Futures endpoints.
    '''

class KrakenFuturesWSClient(KrakenFuturesWSClientCl):
    ''' Class that is used as a client to create a websocket connection
        and handle un/subscribing, reconnecting, ...
    '''



'''Module that provides the Spot REST clients'''
from kraken.spot.market.market import MarketClient
from kraken.spot.trade.trade import TradeClient
from kraken.spot.user.user import UserClient
from kraken.spot.funding.funding import FundingClient
from kraken.spot.staking.staking import StakingClient
from kraken.spot.websocket.websocket import KrakenSpotWSClientCl

class User(UserClient):
    ''' Class that is used as a client to access the user-related
        Kraken Futures endpoints.
    '''
class Trade(TradeClient):
    ''' Class that is used as a client to access the trade-related
        Kraken Futures endpoints.
    '''
class Market(MarketClient):
    ''' Class that is used as a client to access the market-related
        Kraken Futures endpoints.
    '''

class Funding(FundingClient):
    ''' Class that is used as a client to access the funding-related
        Kraken Futures endpoints.
    '''

class Staking(StakingClient):
    ''' Class that is used as a client to access the staking-related
        Kraken Futures endpoints.
    '''

class KrakenSpotWSClient(KrakenSpotWSClientCl):
    ''' Class that is used as a client to create a websocket connection
        and handle un/subscribing, reconnecting, ...
    '''



'''Module that implements the Kraken Spot Stakung client'''
from kraken.base_api.base_api import KrakenBaseRestAPI
class StakingClient(KrakenBaseRestAPI):
    '''Class that implements the Kraken Spot Stakung client'''

    def stake_asset(self, asset: str, amount: str, method: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/stake'''
        return self._request(
            method='POST',
            uri='/private/Stake',
            params={
                'asset': asset,
                'amount': amount,
                'method': method
            },
            auth=True
        )

    def unstake_asset(self, asset: str, amount: str, method: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/unstake'''
        return self._request(
            method='POST',
            uri='/private/Unstake',
            params={
                'asset': asset,
                'amount': amount,
                'method': method
            }, auth=True
        )

    def list_stakeable_assets(self) -> dict:
        '''https://docs.kraken.com/rest/#operation/getStakingAssetInfo'''
        return self._request(
            method='POST',
            uri='/private/Staking/Assets',
            auth=True
        )

    def get_pending_staking_transactions(self) -> dict:
        '''https://docs.kraken.com/rest/#operation/getStakingPendingDeposits'''
        return self._request(
            method='POST',
            uri='/private/Staking/Pending',
            auth=True
        )

    def list_staking_transactions(self) -> dict:
        '''https://docs.kraken.com/rest/#operation/getStakingTransactions'''
        return self._request(
            method='POST',
            uri='/private/Staking/Transactions',
            auth=True
        )



'''Module that provides custom exceptions for the python-kraken-sdk'''
import functools
from typing import List
class KrakenExceptions():
    '''https://docs.kraken.com/rest/#section/General-Usage/Requests-Responses-and-Errors'''

    def __init__(self, msg=None, *args, **kwargs):
        self.EXCEPTION_ASSIGNMENT = {

            # spot/margin_trading_rrors______________________
            'EGeneral:Invalid arguments': self.KrakenInvalidArgumentsError,
            'EGeneral:Invalid arguments:Index unavailable': self.KrakenInvalidArgumentsIndexUnavailableError,
            'EGeneral:Permission denied': self.KrakenPermissionDeniedError,

            'EService:Unavailable': self.KrakenServiceUnavailableError,
            'EService:Market in cancel_only mode': self.KrakenMarketInOnlyCancelModeError,
            'EService:Market in post_only mode': self.KrakenMarketInOnlyPostModeError,
            'EService:Deadline elapsed': self.KrakenDeadlineElapsedError,

            'EAPI:Invalid key': self.KrakenInvalidAPIKeyError,
            'EAPI:Invalid signature': self.KrakenInvalidSignatureError,
            'EAPI:Invalid nonce': self.KrakenInvalidNonceError,

            'EOrder:Invalid order': self.KrakenInvalidOrderError,
            'EOrder:Cannot open position': self.KrakenCannotOpenPositionError,
            'EOrder:Margin allowance exceeded': self.KrakenMarginAllowedExceededError,
            'EOrder:Margin level too low': self.KrakenMarginLevelToLowError,
            'EOrder:Margin position size exceeded': self.KrakenMarginPositionSizeExceededError,
            'EOrder:Insufficient margin': self.KrakenInsufficientMarginError,
            'EOrder:Insufficient funds': self.KrakenInsufficientFundsError,
            'EOrder:Order minimum not met': self.KrakenOrderMinimumNotMetError,
            'EOrder:Cost minimum not met': self.KrakenCostMinimumNotMetError,
            'EOrder:Tick size check failed': self.KrakenTickSizeInvalidCheckError,
            'EOrder:Orders limit exceeded': self.KrakenOrderLimitsExceededError,
            'EOrder:Rate limit exceeded': self.KrakenRateLimitExceededError,
            'EOrder:Positions limit exceeded': self.KrakenPositionLimitExceeded,
            'EOrder:Unknown position': self.KrakenUnknownPositionError,

            'EFunding:Invalid reference id': self.KrakenInvalidReferenceIdError,
            'EFunding:Unknown withdraw key': self.KrakenUnknownWithdrawKeyError,
            'EFunding:Invalid amount': self.KrakenInvalidAmountError,
            'EFunding:Invalid staking method': self.KrakenInvalidStakingMethodError,
            'EFunding:Too many addresses': self.KrakenToManyAdressesError,
            'EFunding:Unknown asset': self.KrakenUnknownAssetError,

            'EQuery:Unknown asset': self.KrakenUnknownAssetError,
            'EQuery:Unknown asset pair': self.KrakenUnknownAssetPairError,

            # futures_trading_errors__________________________
            'authenticationError': self.KrakenAuthenticationError,
            'insufficientAvailableFunds': self.KrakenInsufficientAvailavleFundsError,
            'apiLimitExceeded': self.KrakenApiLimitExceededError,
            'invalidUnit': self.KrakenInvalidUnitError,
            'Unavailable': self.KrakenUnavailableError,
            'invalidAccount': self.KrakenInvalidAccountError,
            'notFound': self.KrakenNotFoundError,
            'orderForEditNotFound': self.KrakenOrderForEditNotFoundError,
        }

    def get_exception(self, data: List[str]):
        '''Returns the exception given by name if available'''
        for name, exception in self.EXCEPTION_ASSIGNMENT.items():
            if data == name: return exception
            if isinstance(data, list):
                if name in data: return exception
                for err in data:
                    if name in err: return exception
        return None

    def docstring_message(cls):
        '''Decorates an exception to make its docstring its default message.
            https://stackoverflow.com/a/66491013/13618168
        '''
        cls_init = cls.__init__
        @functools.wraps(cls.__init__)
        def wrapped_init(self, msg=None, *args, **kwargs):
            err_message = self.__doc__ if not msg else f'{self.__doc__}\ndetails: {msg}'
            cls_init(self, err_message, *args, **kwargs)
        cls.__init__ = wrapped_init
        return cls

    @docstring_message
    class KrakenInvalidArgumentsError(Exception):
        '''The request payload is malformed, incorrect or ambiguous.'''

    @docstring_message
    class KrakenInvalidArgumentsIndexUnavailableError(Exception):
        '''Index pricing is unavailable for stop/profit orders on this pair.'''

    @docstring_message
    class KrakenPermissionDeniedError(Exception):
        '''API key doesn't have permission to make this request.'''

    @docstring_message
    class KrakenServiceUnavailableError(Exception):
        '''The matching engine or API is offline'''

    @docstring_message
    class KrakenMarketInOnlyCancelModeError(Exception):
        '''Request can't be made at this time. Please check system status.'''

    @docstring_message
    class KrakenMarketInOnlyPostModeError(Exception):
        '''Request can't be made at this time. Please check system status.'''

    @docstring_message
    class KrakenDeadlineElapsedError(Exception):
        '''The request timed out according to the default or specified deadline'''

    @docstring_message
    class KrakenInvalidAPIKeyError(Exception):
        '''An invalid API-Key header was supplied.'''

    @docstring_message
    class KrakenInvalidSignatureError(Exception):
        '''An invalid API-Sign header was supplied.'''

    @docstring_message
    class KrakenInvalidNonceError(Exception):
        '''An invalid nonce was supplied'''

    @docstring_message
    class KrakenInvalidOrderError(Exception):
        '''Order is invalid.'''

    @docstring_message
    class KrakenAuthenticationError(Exception):
        '''Credentials are invalid.'''

    @docstring_message
    class KrakenCannotOpenPositionError(Exception):
        '''User/tier is ineligible for margin trading.'''

    @docstring_message
    class KrakenMarginAllowedExceededError(Exception):
        '''User has exceeded their margin allowance.'''

    @docstring_message
    class KrakenMarginLevelToLowError(Exception):
        '''Client has insufficient equity or collateral.'''

    @docstring_message
    class KrakenMarginPositionSizeExceededError(Exception):
        '''Client would exceed the maximum position size for this pair.'''

    @docstring_message
    class KrakenInsufficientMarginError(Exception):
        '''Exchange does not have available funds for this margin trade.'''

    @docstring_message
    class KrakenInsufficientFundsError(Exception):
        '''Client does not have the necessary funds.'''

    @docstring_message
    class KrakenInsufficientAvailavleFundsError(Exception):
        '''Client does not have the necessary funds.'''

    @docstring_message
    class KrakenOrderMinimumNotMetError(Exception):
        '''Order size does not meet ordermin.'''

    @docstring_message
    class KrakenCostMinimumNotMetError(Exception):
        '''Cost (price * volume) does not meet costmin.'''

    @docstring_message
    class KrakenTickSizeInvalidCheckError(Exception):
        '''Price submitted is not a valid multiple of the pair's tick_size.'''

    @docstring_message
    class KrakenOrderLimitsExceededError(Exception):
        '''Order limits exceeded. Please check your open orders limit.'''

    @docstring_message
    class KrakenRateLimitExceededError(Exception):
        '''API rate limit exceeded. Please check your rate limits.'''

    @docstring_message
    class KrakenApiLimitExceededError(Exception):
        '''API rate limit exceeded. Please check your rate limits.'''

    @docstring_message
    class KrakenPositionLimitExceeded(Exception):
        '''Position limit exceeded. Please check your limits.'''

    @docstring_message
    class KrakenUnknownPositionError(Exception):
        '''Position is unknown.'''

    @docstring_message
    class KrakenUnknownAssetPairError(Exception):
        '''The asset pair is unknown.'''

    @docstring_message
    class KrakenUnknownAssetError(Exception):
        '''The asset is unknown'''

    @docstring_message
    class KrakenInvalidUnitError(Exception):
        '''The specified unit is invalid.'''

    @docstring_message
    class KrakenUnavailableError(Exception):
        '''The requested resource is unavailable.'''

    @docstring_message
    class KrakenInvalidReferenceIdError(Exception):
        '''The requested referece id is invalid.'''

    @docstring_message
    class KrakenUnknownWithdrawKeyError(Exception):
        '''The requested withdrawal key is unknown.'''

    @docstring_message
    class KrakenInvalidAmountError(Exception):
        '''The specified amount is invalid.'''

    @docstring_message
    class KrakenInvalidStakingMethodError(Exception):
        '''The staking method is invalid.'''

    @docstring_message
    class KrakenInvalidAccountError(Exception):
        '''The account is invalid.'''

    @docstring_message
    class KrakenNotFoundError(Exception):
        '''The resource is not found.'''

    @docstring_message
    class KrakenOrderForEditNotFoundError(Exception):
        '''The order for edit could not be found.'''

    @docstring_message
    class KrakenToManyAdressesError(Exception):
        '''To many adresses specified.'''
    # ? ____CUSTOM_EXCEPTIONS_________
    @docstring_message
    class MaxReconnectError(Exception):
        '''To many reconnect tries.'''



'''Module that implements the Kraken Spot User client'''
from kraken.base_api.base_api import KrakenBaseRestAPI

class UserClient(KrakenBaseRestAPI):
    '''Class that implements the Kraken Spot User client'''

    def get_account_balance(self) -> dict:
        '''https://docs.kraken.com/rest/#operation/getAccountBalance'''
        return self._request(method='POST', uri='/private/Balance')

    def get_balances(self, currency: str) -> dict:
        '''Returns the balance and available balance of a given currency'''

        balance = float(0)
        currency_found = False
        for symbol, value in self.get_account_balance().items():
            if balance != float(0): break
            if symbol in [ currency, f'Z{currency}', f'X{currency}' ]:
                balance = float(value)
                currency_found = True
        if not currency_found: raise ValueError('Currency not found!')

        available_balance = balance
        for _, order in self.get_open_orders()['open'].items():
            if currency in order['descr']['pair'][0:len(currency)]:
                if order['descr']['type'] == 'sell':
                    available_balance -= float(order['vol'])
            elif currency in order['descr']['pair'][-len(currency):]:
                if order['descr']['type'] == 'buy':
                    available_balance -= float(order['vol']) * float(order['descr']['price'])

        return {
            'currency': currency,
            'balance': balance,
            'available_balance': available_balance
        }

    def get_trade_balance(self, asset: str=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradeBalance'''
        params = {}
        if asset is not None: params['asset'] = asset
        return self._request(method='POST', uri='/private/TradeBalance', params=params)

    def get_open_orders(self, trades: bool=False, userref: int=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getOpenOrders'''
        params = { 'trades': trades }
        if userref is not None: params['userref'] = userref
        return self._request(method='POST', uri='/private/OpenOrders', params=params)

    def get_closed_orders(self,
        trades: bool=False,
        userref: int=None,
        start: int=None,
        end: int=None,
        ofs: int=None,
        closetime: str='both'
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/getClosedOrders'''
        params = {
            'trades': trades,
            'closetime': closetime
        }
        if userref is not None: params['userref'] = userref
        if start is not None: params['start'] = start
        if end is not None: params['end'] = end
        if ofs is not None: params['ofs'] = ofs

        return self._request(method='POST', uri='/private/ClosedOrders', params=params)

    def get_orders_info(self, txid, trades: bool=False, userref: int=None) -> dict:
        '''https://docs.kraken.com/rest/#tag/User-Data/operation/getOrdersInfo'''
        params = {
            'txid': txid,
            'trades': trades
        }
        if isinstance(txid, list): params['txid'] = self._to_str_list(txid)
        if userref is not None: params['userref'] = userref
        return self._request(method='POST', uri='/private/QueryOrders', params=params)

    def get_trades_history(self,
        type_: str='all',
        trades: bool=False,
        start: int=None,
        end: int=None,
        ofs: int=None
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradeHistory'''
        params = {
            'type': type_,
            'trades': trades
        }
        if start is not None: params['start'] = start
        if end is not None: params['end'] = end
        if ofs is not None: params['ofs'] = ofs
        return self._request(method='POST', uri='/private/TradesHistory', params=params)

    def get_trades_info(self, txid: str, trades: bool=False) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradesInfo'''
        return self._request(method='POST', uri='/private/QueryTrades', params={
            'trades': trades,
            'txid': self._to_str_list(txid)
        })

    def get_open_positions(self, txid=None, docalcs: bool=False, consolidation: str='market') -> dict:
        '''https://docs.kraken.com/rest/#operation/getOpenPositions'''
        params = {
            'docalcs': docalcs,
            'consolidation': consolidation
        }
        if txid is not None: params['txid'] = self._to_str_list(txid)
        return self._request(method='POST', uri='/private/OpenPositions', params=params)

    def get_ledgers_info(self,
        asset: str='all',
        aclass: str='currency',
        type_: str='all',
        start: int=None,
        end: int=None,
        ofs: int=None
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/getLedgers'''
        params = {
            'asset': asset,
            'aclass': aclass,
            'type': type_
        }
        if isinstance(params['asset'], list): params['asset'] = self._to_str_list(asset)
        if start is not None: params['start'] = start
        if end is not None: params['end'] = end
        if ofs is not None: params['ofs'] = ofs
        return self._request(method='POST', uri='/private/Ledgers', params=params)

    def get_ledgers(self, id_, trades: bool=False) -> dict:
        '''https://docs.kraken.com/rest/#operation/getLedgersInfo'''
        return self._request(method='POST', uri='/private/QueryLedgers', params={
            'trades': trades,
            'id': self._to_str_list(id_)
        })

    def get_trade_volume(self, pair=None, fee_info: bool=True) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradeVolume'''
        params = { 'fee-info': fee_info}
        if pair is not None: params['pair'] = self._to_str_list(pair)
        return self._request(method='POST', uri='/private/TradeVolume', params=params)

    def request_export_report(self,
        report: str,
        description: str,
        format_: str='CSV',
        fields: str='all',
        starttm: int=None,
        endtm: int=None,
        **kwargs
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/addExport

        ---- RESPONSE ----
            {'id': 'INSG'}
        '''
        if report not in ['trades', 'ledgers']:
            raise ValueError('report must be one of "trades", "ledgers"')
        params = {
            'report': report,
            'description': description,
            'format': format_,
            'fields': self._to_str_list(fields)
        }
        params.update(kwargs)
        if starttm is not None: params['starttm'] = starttm
        if endtm is not None: params['endtm'] = endtm
        return self._request(method='POST', uri='/private/AddExport', params=params)

    def get_export_report_status(self, report: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/exportStatus

        ---- RESPONSE ----
        [{'id': 'INSG', 'descr': 'myLedgers1', 'format': 'CSV', 'report': 'ledgers', 'status': 'Processed', 'aclass': 'currency', 'fields': 'all', 'asset': 'all', 'subtype': 'all', 'starttm': '1656633600', 'endtm': '1657355882', 'createdtm': '1657355882', 'expiretm': '1658565482', 'completedtm': '1657355887', 'datastarttm': '1656633600', 'dataendtm': '1657355882', 'flags': '0'}]
        '''
        if report not in ['trades', 'ledgers']:
            raise ValueError('report must be one of "trades", "ledgers"')
        return self._request(
            method='POST',
            uri='/private/ExportStatus',
            params={ 'report': report }
        )

    def retrieve_export(self, id_: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/retrieveExport'''
        return self._request(
            method='POST',
            uri='/private/RetrieveExport',
            params={ 'id': id_ },
            return_raw=True
        )

    def delete_export_report(self, id_: str, type_: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/removeExport'''
        return self._request(
            method='POST',
            uri='/private/RemoveExport',
            params={ 'id': id_, 'type': type_ }
        )



'''Module that implements the Kraken Futures Funding client'''
from kraken.base_api.base_api import KrakenBaseFuturesAPI

class FundingClient(KrakenBaseFuturesAPI):
    '''Class that implements the Kraken Futures Funding client'''

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False) -> None:
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)

    def get_historical_funding_rates(self, symbol: str) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-historical-funding-rates-historicalfundingrates'''
        return self._request(
            method='GET',
            uri='/derivatives/api/v4/historicalfundingrates',
            query_params={'symbol': symbol},
            auth=False
        )

    def initiate_wallet_transfer(self, amount: str, fromAccount: str, toAccount: str, unit: str) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-transfers-initiate-wallet-transfer'''
        return self._request(
            method='POST', uri='/derivatives/api/v3/transfer',
            post_params={
                'amount': amount,
                'fromAccount': fromAccount,
                'toAccount': toAccount,
                'unit': unit
            },
            auth=True
        )

    def initiate_subccount_transfer(self,
        amount: str,
        fromAccount: str,
        fromUser: str,
        toAccount: str,
        toUser: str,
        unit: str
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-transfers-initiate-sub-account-transfer'''
        return self._request(
            method='POST',
            uri='/derivatives/api/v3/transfer/subaccount',
            post_params={
                'amount': amount,
                'fromAccount': fromAccount,
                'fromUser': fromUser,
                'toAccount': toAccount,
                'toUser': toUser,
                'unit': unit
            },
            auth=True
        )

    def initiate_withdrawal_to_spot_wallet(self, amount: str, currency: str, sourceWallet: str=None, **kwargs) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-transfers-initiate-withdrawal-to-spot-wallet'''
        if self.sandbox: raise ValueError('This function is not available in sandbox mode.')
        params = {
            'amount': str(amount),
            'currency': currency,
        }
        if sourceWallet is not None: params['sourceWallet'] = sourceWallet
        params.update(kwargs)
        return self._request(method='POST', uri='/derivatives/api/v3/withdrawal', post_params=params, auth=True)



'''Module that implements the Kraken Futures websocket client'''
import sys
import logging
import asyncio
from typing import List
from random import random
import traceback
import json
import copy
import websockets
try:
    from kraken.futures.ws_client.ws_client import FuturesWsClientCl
    from kraken.exceptions.exceptions import KrakenExceptions
except ModuleNotFoundError:
    print('USING LOCAL MODULE')
    sys.path.append('/Users/benjamin/repositories/Trading/python-kraken-sdk')
    from kraken.futures.ws_client.ws_client import FuturesWsClientCl
    from kraken.exceptions.exceptions import KrakenExceptions


class ConnectFuturesWebsocket():
    '''
        This class is only called by the KrakenFuturesWSClientCl class
        to establish and handle a websocket connection.

        ====== P A R A M E T E R S ======
        client: kraken.futures.client.KrakenFuturesWSClient
            the websocket client
        endpoint: str
            endpoint/url to connect with
        callback: function [optional], default=None
            callback function to call when a message is received
    '''

    MAX_RECONNECT_NUM = 2

    def __init__(self, client, endpoint: str, callback):
        self.__client = client
        self.__ws_endpoint = endpoint
        self.__callback = callback

        self.__reconnect_num = 0

        self.__last_challenge = None
        self.__new_challenge = None
        self.__challenge_ready = False

        self.__socket = None
        self.__subscriptions = []

        asyncio.ensure_future(
            self.__run_forever(),
            loop=asyncio.get_running_loop()
        )

    @property
    def subscriptions(self) -> list:
        '''Returns the active subscriptions'''
        return self.__subscriptions

    async def __run(self, event: asyncio.Event):
        keep_alive = True
        self.__new_challenge = None
        self.__last_challenge = None

        async with websockets.connect(f'wss://{self.__ws_endpoint}', ping_interval=30) as socket:
            logging.info('Websocket connected!')
            self.__socket = socket

            if not event.is_set():
                event.set()
            self.__reconnect_num = 0

            while keep_alive:
                try: _msg = await asyncio.wait_for(self.__socket.recv(), timeout=15)
                except asyncio.TimeoutError: pass # important
                except asyncio.CancelledError:
                    logging.exception('asyncio.CancelledError')
                    keep_alive = False
                    await self.__callback({'error': 'asyncio.CancelledError'})
                else:
                    try: msg = json.loads(_msg)
                    except ValueError: logging.warning(_msg)
                    else:
                        forward = True
                        if 'event' in msg:
                            _event = msg['event']
                            if _event == 'challenge' and 'message' in msg:
                                forward = False
                                self.__handle_new_challenge(msg)
                            elif _event == 'subscribed': self.__append_subscription(msg)
                            elif _event == 'unsubscribed': self.__remove_subscription(msg)
                        if forward: await self.__callback(msg)

    async def __run_forever(self) -> None:
        try:
            while True: await self.__reconnect()
        except KrakenExceptions.MaxReconnectError:
            await self.__callback({
                'error': 'kraken.exceptions.exceptions.KrakenExceptions.MaxReconnectError'
            })
        except Exception:
            # for task in asyncio.all_tasks(): task.cancel()
            logging.error(traceback.format_exc())
        # except asyncio.CancelledError: pass
        finally: self.__client.exception_occur = True

    async def __reconnect(self):
        logging.info('Websocket start connect/reconnect')

        self.__reconnect_num += 1
        if self.__reconnect_num >= self.MAX_RECONNECT_NUM:
            raise KrakenExceptions.MaxReconnectError()

        reconnect_wait = self.__get_reconnect_wait(self.__reconnect_num)
        logging.debug(
            f'asyncio sleep reconnect_wait={reconnect_wait} s reconnect_num={self.__reconnect_num}'
        )
        await asyncio.sleep(reconnect_wait)
        logging.debug('asyncio sleep done')
        event = asyncio.Event()

        tasks = {
            asyncio.ensure_future(
                self.__recover_subscription_req_msg(event)
            ): self.__recover_subscription_req_msg,
            asyncio.ensure_future(self.__run(event)): self.__run
        }

        while set(tasks.keys()):
            finished, pending = await asyncio.wait(
                tasks.keys(),
                return_when=asyncio.FIRST_EXCEPTION
            )
            exception_occur = False
            for task in finished:
                if task.exception():
                    exception_occur = True
                    message = f'{task} got an exception {task.exception()}'
                    message += f'\nTRACEBACK: {traceback.format_exc()}'
                    logging.warning(message)
                    for process in pending:
                        logging.warning(f'pending {process}')
                        try: process.cancel()
                        except asyncio.CancelledError: logging.exception('CancelledError')
                        logging.warning('cancel ok')
                    await self.__callback({ 'error': message })
            if exception_occur: break
        logging.warning('reconnect over')

    async def __recover_subscription_req_msg(self, event) -> None:
        logging.info(f'Recover subscriptions {self.__subscriptions} waiting.')
        await event.wait()

        for sub in self.__subscriptions:
            if sub['feed'] in self.__client.get_available_private_subscription_feeds():
                await self.send_message(copy.deepcopy(sub), private=True)
            elif sub['feed'] in self.__client.get_available_public_subscription_feeds():
                await self.send_message(copy.deepcopy(sub), private=False)
            logging.info(f'{sub}: OK')

        logging.info(f'Recover subscriptions {self.__subscriptions} done.')

    async def send_message(self, msg: dict, private: bool=False) -> None:
        '''Sends a message via the websocket connection'''
        while not self.__socket:
            await asyncio.sleep(.4)

        if private:
            if not self.__client.is_auth:
                raise ValueError('Cannot access private endpoints with unauthenticated client!')
            if not self.__challenge_ready:
                await self.__check_challenge_ready()

            msg['api_key'] = self.__client._key
            msg['original_challenge'] = self.__last_challenge
            msg['signed_challenge'] = self.__new_challenge

        await self.__socket.send(json.dumps(msg))

    def __handle_new_challenge(self, msg: dict) -> None:
        self.__last_challenge = msg['message']
        self.__new_challenge = self.__client._get_sign_challenge(self.__last_challenge)
        self.__challenge_ready = True

    async def __check_challenge_ready(self) -> None:
        await self.__socket.send(json.dumps({
            'event': 'challenge',
            'api_key': self.__client._key
        }))

        logging.debug('Awaiting challenge...')
        while not self.__challenge_ready: await asyncio.sleep(.2)

    def __get_reconnect_wait(self, attempts: int) -> float:
        return round(random() * min(60*3, (2 ** attempts) - 1) + 1)

    def __append_subscription(self, msg: dict) -> None:
        self.__remove_subscription(msg=msg) # remove from list, to avoid duplicates
        sub = self.__build_subscription(msg)
        self.__subscriptions.append(sub)

    def __remove_subscription(self, msg: dict) -> None:
        sub = self.__build_subscription(msg)
        self.__subscriptions = [x for x in self.__subscriptions if x != sub]

    def __build_subscription(self, subscription: dict) -> dict:
        sub = { 'event': 'subscribe' }

        if 'event' not in subscription                                        \
            or subscription['event'] not in ['subscribed', 'unsubscribed']    \
            or 'feed' not in subscription:
            raise ValueError('Cannot append/remove subscription with missing attributes.')

        if subscription['feed'] in self.__client.get_available_public_subscription_feeds():
            # public subscribe
            if 'product_ids' in subscription:
                if isinstance(subscription['product_ids'], list):
                    sub['product_ids'] = subscription['product_ids']
                else:
                    sub['product_ids'] = [subscription['product_ids']]
            sub['feed'] = subscription['feed']

        elif subscription['feed'] in self.__client.get_available_private_subscription_feeds():
            # private subscription
            sub['feed'] = subscription['feed']
        else: logging.warning(
            'Feed not implemented. Please contact the python-kraken-sdk package author.'
        )
        return sub

    def get_active_subscriptions(self) -> List[dict]:
        '''Returns the active subscriptions'''
        return self.__subscriptions


class KrakenFuturesWSClientCl(FuturesWsClientCl):
    '''https://docs.futures.kraken.com/#websocket-api

        Class to access public and (optional)
        private/authenticated websocket connection.

        ====== P A R A M E T E R S ======
        key: str, [optional], default: ''
            API Key for the Kraken API
        secret: str, [optional], default: ''
            Secret API Key for the Kraken API
        url: str, [optional], default: ''
            Set a specific/custom url
        callback: async function [optional], default=None
            callback function which receives the websocket messages
        sandbox: bool [optional], default=False
            use the Kraken Futures demo url

        ====== E X A M P L E ======
        import asyncio
        from kraken.futures.client import KrakenFuturesWSClient

        async def main() -> None:

            # ___Custom_Trading_Bot__________
            class Bot(KrakenFuturesWSClient):

                async def on_message(self, event) -> None:
                    print(event)

            bot = Bot() # unauthenticated
            auth_bot = Bot(key=key, secret=secret) # authenticated

            # ... now call for example subscribe and so on

            while True: await asyncio.sleep(6)

        if __name__ == '__main__':
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try: asyncio.run(main())
            except KeyboardInterrupt: loop.close()
    '''

    PROD_ENV_URL = 'futures.kraken.com/ws/v1'
    DEMO_ENV_URL = 'demo-futures.kraken.com/ws/v1'

    def __init__(self,
        key: str='',
        secret: str='',
        url: str='',
        callback=None,
        sandbox: bool=False
    ):
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)

        self.exception_occur = False
        self.__callback = callback
        self._conn = ConnectFuturesWebsocket(
            client=self,
            endpoint= url if url != '' else self.DEMO_ENV_URL if sandbox else self.PROD_ENV_URL,
            callback=self.on_message
        )

    async def on_message(self, msg) -> None:
        ''' Calls the defined callback function (if defined)
            or overload this function

            ====== P A R A M E T E R S ======
            msg: dict
                message received from Kraken via the websocket connection

            ====== N O T E S ======
            Can be overloaded like in the documentation of this class.
        '''
        if self.__callback is not None:
            await self.__callback(msg)
        else:
            logging.warning('Received event but no callback is defined')
            logging.info(msg)

    async def subscribe(self, feed: str, products: List[str]=None) -> None:
        '''Subscribe to a channel/feed
            https://docs.futures.kraken.com/#websocket-api-websocket-api-introduction-subscriptions

            ====== P A R A M E T E R S ======
            subscription: dict
                the subscription to subscribe to
            products: List[str]
                list of assets or list of a single product

            ====== E X A M P L E ======
            # ... initialize bot as documented on top of this class.
            await bot.subscribe(feed='ticker', products=["XBTUSD", "DOT/EUR"])

            ====== N O T E S ======
            Success or failures are sent over the websocket connection and can be
            received via the on_message callback function.
        '''

        message = { 'event': 'subscribe', 'feed': feed }

        if products is not None:
            if not isinstance(products, list):
                raise ValueError(
                    'Parameter products must be type of List[str] (e.g. pair=["PI_XBTUSD"])'
                )
            message['product_ids'] = products

        if feed in self.get_available_private_subscription_feeds():
            if products is not None:
                raise ValueError('There is no private feed that accepts products!')
            await self._conn.send_message(message, private=True)
        elif feed in self.get_available_public_subscription_feeds():
            if products is not None:
                for product in products:
                    sub = copy.deepcopy(message)
                    sub['product_ids'] = [product]
                    await self._conn.send_message(sub, private=False)
            else: await self._conn.send_message(message, private=False)
        else: raise ValueError(f'Feed: {feed} not found. Not subscribing to it.')

    async def unsubscribe(self, feed: str, products: List[str]=None) -> None:
        '''Unsubscribe from a topic/feed
            https://docs.futures.kraken.com/#websocket-api-websocket-api-introduction-subscriptions

            ====== P A R A M E T E R S ======
            subscription: dict
                the subscription to unsubscribe from
            products: List[str]
                list of assets or list of a single product

            ====== E X A M P L E ======
            # ... initialize bot as documented on top of this class.
            bot.unsubscribe(feed='ticker', products=["XBTUSD", "DOT/EUR"])

            ====== N O T E S ======
            Success or failures are sent over the websocket connection and can be
            received via the on_message callback function.
        '''

        message = { 'event': 'unsubscribe', 'feed': feed }

        if products is not None:
            if not isinstance(products, list):
                raise ValueError('Parameter products must be type of List[str]\
                    (e.g. pair=["PI_XBTUSD"])')
            message['product_ids'] = products

        if feed in self.get_available_private_subscription_feeds():
            if products is not None:
                raise ValueError('There is no private feed that accepts products!')
            await self._conn.send_message(message, private=True)
        elif feed in self.get_available_public_subscription_feeds():
            if products is not None:
                for product in products:
                    sub = copy.deepcopy(message)
                    sub['product_ids'] = [product]
                    await self._conn.send_message(sub, private=False)
            else: await self._conn.send_message(message, private=False)
        else: raise ValueError(f'Feed: {feed} not found. Not unsubscribing it.')

    @staticmethod
    def get_available_public_subscription_feeds() -> List[str]:
        '''Return all available public feeds to subsribe.'''
        return [ 'trade', 'book', 'ticker', 'ticker_lite', 'heartbeat']

    @staticmethod
    def get_available_private_subscription_feeds() -> List[str]:
        '''Return all available private feeds to subsribe.'''
        return [
            'fills',
            'open_positions', 'open_orders', 'open_orders_verbose',
            'balances',
            'deposits_withdrawals', 'account_balances_and_margins',
            'account_log', 'notifications_auth'
        ]

    @property
    def is_auth(self) -> bool:
        '''Checks if key and secret are set.'''
        return self._key and self._secret

    def get_active_subscriptions(self) -> List[dict]:
        '''Returns the acitve subscriptions.'''
        return self._conn.get_active_subscriptions()



#!/usr/bin/python3
# -*- coding:utf-8 -*-
# run via  python3 setup.py upload

import io, os, sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'python-kraken-sdk'
DESCRIPTION = 'Collection of clients and methods to interact with the Kraken cryptocurrency exchange.'
URL = 'https://github.com/btschwertfeger/Python-Kraken-SDK'
EMAIL = 'development@b-schwertfeger.de'
AUTHOR = 'Benjamin Thomas Schwertfeger'
REQUIRES_PYTHON = '>=3.7.0'
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    'requests', 'websockets', 'asyncio>=3.4',
]

# What packages are optional?
EXTRAS = {
    'testing': ['pytest', 'tqdm']
}

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f'\n{f.read()}'
except FileNotFoundError:
    long_description = DESCRIPTION

about = {}
if not VERSION:
    project_slug = 'kraken'
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    raise ValueError('Version not found!')
    exit()

class UploadCommand(Command):
    '''Support setup.py upload.'''

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        print(f'\033[1m{s}\033[0m')

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(f'{sys.executable} setup.py sdist bdist_wheel --universal')

        self.status('Testing the build using flake8')
        if os.system('flake8 . --select=E9,F63,F7,F82 --show-source --statistics') != 0:
            self.status('Testing failed, build has some errors in it!')
            exit(1)

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        # self.status('Pushing git tags…')
        # os.system(f'git tag v{about['__version__']}')
        # os.system('git push --tags')

        sys.exit()

class TestUploadCommand(Command):
    '''Support setup.py test upload.'''

    description = 'Build and test publishing the package.'
    user_options = []

    @staticmethod
    def status(s):
        print(f'\033[1m{s}\033[0m')

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(f'{sys.executable} setup.py sdist bdist_wheel --universal')

        self.status('Testing the build using flake8')
        if os.system('flake8 . --select=E9,F63,F7,F82 --show-source --statistics') != 0:
            self.status('Testing failed, build has some errors in it!')
            sys.exit(1)

        self.status('Uploading the package to test PyPI via Twine…')
        os.system('twine upload -r testpypi dist/*')#--repository-url https://test.pypi.org/legacy/ dist/*')

        sys.exit()
class TestCommand(Command):

    description = 'Build and test the package.'
    user_options = []

    @staticmethod
    def status(s):
        print(f'\033[1m{s}\033[0m')

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(f'{sys.executable} setup.py sdist bdist_wheel --universal')

        self.status('Testing the build')
        if os.system('flake8 . --select=E9,F63,F7,F82 --show-source --statistics') != 0: exit(1)
        print('Success')
        sys.exit()

setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=['tests', '*.tests', '*.tests.*', 'tests.*', '*.env*']),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],

    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='GPLv3',
    classifiers=[
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Development Status :: 4 - Beta',
        'Framework :: AsyncIO',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Unix'
    ],
    cmdclass={
        'upload': UploadCommand,
        'test': TestCommand,
        'testupload': TestUploadCommand,
    },
)



'''Module that implements the kraken Spot websocket clients'''
import sys
import logging
import json
import time
import asyncio
from random import random
import traceback
import copy
from typing import List
import websockets

try:
    from kraken.spot.ws_client.ws_client import SpotWsClientCl
    from kraken.exceptions.exceptions import KrakenExceptions
except ModuleNotFoundError:
    print('USING LOCAL MODULE')
    sys.path.append('/Users/benjamin/repositories/Trading/python-kraken-sdk')
    from kraken.spot.client.ws_client import SpotWsClientCl
    from kraken.exceptions.exceptions import KrakenExceptions

class ConnectSpotWebsocket:
    '''
        This class is only called by the KrakenSpotWSClientCl class
        to establish and handle a websocket connection.

        ====== P A R A M E T E R S ======
        client: kraken.spot.client.KrakenSpotWSClient
            the websocket client
        endpoint: str
            endpoint/url to connect with
        callback: function [optional], default=None
            callback function to call when a message is received
        private: bool [optional], default=False
            if client is authenticated to send signed messages
            and get private feeds

    '''

    MAX_RECONNECT_NUM = 10

    def __init__(self, client, endpoint: str, callback, is_auth: bool=False):
        self.__client = client
        self.__ws_endpoint = endpoint
        self.__callback = callback

        self.__reconnect_num = 0
        self.__ws_conn_details = None

        self.__is_auth = is_auth

        self.__last_ping = None
        self.__socket = None
        self.__subscriptions = []

        asyncio.ensure_future(
            self.__run_forever(),
            loop=asyncio.get_running_loop()
        )

    @property
    def subscriptions(self) -> list:
        '''Returns the active subscriptions'''
        return self.__subscriptions

    @property
    def is_auth(self) -> bool:
        '''Returns true if the connection can access privat endpoints'''
        return self.__is_auth

    async def __run(self, event: asyncio.Event):
        keep_alive = True
        self.__last_ping = time.time()
        self.__ws_conn_details = None if not self.__is_auth else self.__client.get_ws_token()
        logging.debug(f'Websocket token: {self.__ws_conn_details}')

        async with websockets.connect(f'wss://{self.__ws_endpoint}', ping_interval=None) as socket:
            logging.info('Websocket connected!')
            self.__socket = socket

            if not event.is_set():
                await self.send_ping()
                event.set()
            self.__reconnect_num = 0

            while keep_alive:
                if time.time() - self.__last_ping > 10: await self.send_ping()
                try:
                    _msg = await asyncio.wait_for(self.__socket.recv(), timeout=15)
                except asyncio.TimeoutError: # important
                    await self.send_ping() 
                except asyncio.CancelledError:
                    logging.exception('asyncio.CancelledError')
                    keep_alive = False
                    await self.__callback({'error': 'asyncio.CancelledError'})
                else:
                    try: msg = json.loads(_msg)
                    except ValueError: logging.warning(_msg)
                    else:
                        if 'event' in msg:
                            if msg['event'] == 'subscriptionStatus' and 'status' in msg:
                                # remove/assign un/subscriptions
                                try:
                                    if msg['status'] == 'subscribed':
                                        self.__append_subscription(msg)
                                    elif msg['status'] == 'unsubscribed':
                                        self.__remove_subscription(msg)
                                    elif msg['status'] == 'error': logging.warning(msg)
                                except AttributeError: pass
                        await self.__callback(msg)

    async def __run_forever(self) -> None:
        try:
            while True: await self.__reconnect()
        except KrakenExceptions.MaxReconnectError:
            await self.__callback({'error': 'kraken.exceptions.exceptions.KrakenExceptions.MaxReconnectError'})
        except Exception as execption:
            logging.error(f'{execption}: {traceback.format_exc()}')
        finally: self.__client.exception_occur = True

    async def __reconnect(self):
        logging.info('Websocket start connect/reconnect')

        self.__reconnect_num += 1
        if self.__reconnect_num >= self.MAX_RECONNECT_NUM:
            raise KrakenExceptions.MaxReconnectError()

        reconnect_wait = self.__get_reconnect_wait(self.__reconnect_num)
        logging.debug(
            f'asyncio sleep reconnect_wait={reconnect_wait} s reconnect_num={self.__reconnect_num}'
        )
        await asyncio.sleep(reconnect_wait)
        logging.debug('asyncio sleep done')
        event = asyncio.Event()

        tasks = {
            asyncio.ensure_future(self.__recover_subscriptions(event)): self.__recover_subscriptions,
            asyncio.ensure_future(self.__run(event)): self.__run
        }

        while set(tasks.keys()):
            finished, pending = await asyncio.wait(tasks.keys(), return_when=asyncio.FIRST_EXCEPTION)
            exception_occur = False
            for task in finished:
                if task.exception():
                    exception_occur = True
                    message = f'{task} got an exception {task.exception()}\nTRACEBACK: {traceback.format_exc()}'
                    logging.warning(message)
                    for process in pending:
                        logging.warning(f'pending {process}')
                        try: process.cancel()
                        except asyncio.CancelledError: logging.exception('asyncio.CancelledError')
                        logging.warning('Cancel OK')
                    await self.__callback({ 'error': message })
            if exception_occur: break
        logging.warning('reconnect over')

    async def __recover_subscriptions(self, event):
        logging.info(
            f'Recover {"auth" if self.__is_auth else "public"} subscriptions {self.__subscriptions} waiting.'
        )
        await event.wait()

        for sub in self.__subscriptions:
            cpy = copy.deepcopy(sub)
            private = False
            if 'subscription' in sub                \
                and 'name' in sub['subscription']   \
                and sub['subscription']['name'] in self.__client.private_sub_names:
                cpy['subscription']['token'] = self.__ws_conn_details['token']
                private = True
            await self.send_message(cpy, private=private)
            logging.info(f'{sub} OK')

        logging.info(
            f'Recovering {"auth" if self.__is_auth else "public"} subscriptions {self.__subscriptions} done.'
        )

    async def send_ping(self):
        '''Sends ping to Keaken'''
        msg = {
            'event': 'ping',
            'reqid': int(time.time() * 1000),
        }
        await self.__socket.send(json.dumps(msg))
        self.__last_ping = time.time()

    async def send_message(self, msg, private: bool=False):
        '''Sends a message via websocket'''
        while not self.__socket: await asyncio.sleep(.4)

        if private and not self.__is_auth:
            raise ValueError('Cannot send private message with public websocket.')

        msg['reqid'] = int(time.time() * 1000)
        if private and 'subscription' in msg: msg['subscription']['token'] = self.__ws_conn_details['token']
        elif private: msg['token'] = self.__ws_conn_details['token']
        await self.__socket.send(json.dumps(msg))

    def __append_subscription(self, msg: dict) -> None:
        ''' Add a dictionary containing subscription information to list
            This is used to recover when the connection gets interrupted.

            This function should only be called in
            when self.__run receives a msg and the following conditions met:
                'event' in msg                           \
                and msg['event'] == 'subscriptionStatus' \
                and 'status' in msg                      \
                and msg['status'] == 'subscribed'

            ====== P A R A M E T E R S ======
            msg: dict
                must be like: { 'subscription': { 'name': '<placeholder>', 'placeholder': 'placeholder', ... }}
        '''
        self.__remove_subscription(msg) # remove from list, to avoid duplicates
        self.__subscriptions.append(self.__build_subscription(msg))

    def __remove_subscription(self, msg: dict) -> None:
        ''' Remove a dictionary containing subscription information from list.

            This function should only be called in
            when self.__run receives a msg and the following conditions met:
                'event' in msg                           \
                and msg['event'] == 'subscriptionStatus' \
                and 'status' in msg                      \
                and msg['status'] == 'unsubscribed'

            ====== P A R A M E T E R S ======
            msg: dict
                must be like: { 'subscription': { 'name': '<placeholder>', 'placeholder': 'placeholder', ... }}
        '''
        sub = self.__build_subscription(msg)
        self.__subscriptions = [x for x in self.__subscriptions if x != sub]

    def __build_subscription(self, msg: dict) -> dict:
        sub = { 'event': 'subscribe' }

        if not 'subscription' in msg or 'name' not in msg['subscription']:
            raise ValueError('Cannot remove subscription with missing attributes.')
        if msg['subscription']['name'] in self.__client.public_sub_names: # public endpoint
            if 'pair' in msg:
                sub['pair'] = msg['pair'] if isinstance(msg['pair'], list) else [msg['pair']]
            sub['subscription'] = msg['subscription']
        elif msg['subscription']['name'] in self.__client.private_sub_names: # private endpoint
            sub['subscription'] = { 'name': msg['subscription']['name'] }
        else: logging.warning('Feed not implemented. Please contact the python-kraken-sdk package author.')
        return sub

    def __get_reconnect_wait(self, attempts: int) -> float:
        return round(random() * min(60*3, (2 ** attempts) - 1) + 1)

class KrakenSpotWSClientCl(SpotWsClientCl):
    '''https://docs.kraken.com/websockets/#overview

        Class to access public and (optional)
        private/authenticated websocket connection.

        This class holds up to two websocket connections, one private
        and one public.

        ====== P A R A M E T E R S ======
        key: str, [optional], default: ''
            API Key for the Kraken API
        secret: str, [optional], default: ''
            Secret API Key for the Kraken API
        url: str, [optional], default: ''
            Set a specific/custom url
        callback: async function [optional], default=None
            callback function which receives the websocket messages
        beta: bool [optional], default=False
            use the Kraken beta url

        ====== P R O P E R T I E S ======
        public_sub_names: List[str]
            list of available public subscription names
        private_sub_names: List[str]
            list of available private subscription names
        active_public_subscriptions: [dict]
            list of active public subscriptions
        active_private_subscriptions: [dict]
            list of active private subscriptions

        ====== E X A M P L E ======
        import asyncio
        from kraken.spot.client import KrakenSpotWSClient

        async def main() -> None:
            class Bot(KrakenSpotWSClient):

                async def on_message(self, event) -> None:
                    print(event)

            bot = Bot() # unauthenticated
            auth_bot = Bot(key='kraken-api-key', secret='kraken-secret-key') # authenticated

            # ... now call for example subscribe and so on

            while True: await asyncio.sleep(6)

        if __name__ == '__main__':
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try: asyncio.run(main())
            except KeyboardInterrupt: loop.close()
    '''

    PROD_ENV_URL = 'ws.kraken.com'
    AUTH_PROD_ENV_URL = 'ws-auth.kraken.com'
    BETA_ENV_URL = 'beta-ws.kraken.com'
    AUTH_BETA_ENV_URL = 'beta-ws-auth.kraken.com'

    def __init__(self, key: str='', secret: str='', url: str='', callback=None, beta: bool=False):
        super().__init__(key=key, secret=secret, url=url, sandbox=beta)
        self.__callback = callback
        self.__is_auth = key and secret

        self.exception_occur = False
        self._pub_conn = ConnectSpotWebsocket(
            client=self,
            endpoint=self.PROD_ENV_URL if not beta else self.BETA_ENV_URL,
            is_auth=False,
            callback=self.on_message
        )

        self._priv_conn = ConnectSpotWebsocket(
             client=self,
             endpoint=self.AUTH_PROD_ENV_URL if not beta else self.AUTH_BETA_ENV_URL,
             is_auth=True,
             callback=self.on_message
        ) if self.__is_auth else None

    async def on_message(self, msg: dict):
        ''' Calls the defined callback function (if defined)
            or overload this function

            ====== P A R A M E T E R S ======
            msg: dict
                message received from Kraken via the websocket connection

            ====== N O T E S ======
            Can be overloaded like in the documentation of this class.
        '''
        if self.__callback is not None: await self.__callback(msg)
        else:
            logging.warning('Received event but no callback is defined')
            print(msg)

    async def subscribe(self, subscription: dict, pair: List[str]=None) -> None:
        '''Subscribe to a channel
            https://docs.kraken.com/websockets-beta/#message-subscribe

            ====== P A R A M E T E R S ======
            subscription: dict
                the subscription to subscribe to
            pair: List[str]
                list of asset pairs or list of a single pair

            ====== E X A M P L E ======
            # ... initialize bot as documented on top of this class.
            await bot.subscribe(subscription={"name": ticker}, pair=["XBTUSD", "DOT/EUR"])

            ====== N O T E S ======
            Success or failures are sent over the websocket connection and can be
            received via the on_message callback function.
        '''

        if 'name' not in subscription: raise AttributeError('Subscription requires a "name" key."')
        private = bool(subscription['name'] in self.private_sub_names)

        payload = {
            'event': 'subscribe',
            'subscription': subscription
        }
        if pair is not None:
            if not isinstance(pair, list): raise ValueError('Parameter pair must be type of List[str] (e.g. pair=["XBTUSD"])')
            payload['pair'] = pair

        if private: # private == without pair
            if not self.__is_auth: raise ValueError('Cannot subscribe to private feeds without valid credentials!')
            if pair is not None: raise ValueError('Cannot subscribe to private endpoint with specific pair!')
            await self._priv_conn.send_message(payload, private=True)

        elif pair is not None: # public with pair
            for symbol in pair:
                sub = copy.deepcopy(payload)
                sub['pair'] = [symbol]
                await self._pub_conn.send_message(sub, private=False)

        else: await self._pub_conn.send_message(payload, private=False)

    async def unsubscribe(self, subscription: dict, pair: List[str]=None) -> None:
        '''Unsubscribe from a topic
            https://docs.kraken.com/websockets/#message-unsubscribe

            ====== P A R A M E T E R S ======
            subscription: dict
                the subscription to unsubscribe from
            pair: List[str]
                list of asset pairs or list of a single pair

            ====== E X A M P L E ======
            # ... initialize bot as documented on top of this class.
            bot.unsubscribe(subscription={"name": ticker}, pair=["XBTUSD", "DOT/EUR"])

            ====== N O T E S ======
            Success or failures are sent over the websocket connection and can be
            received via the on_message callback function.
        '''
        if 'name' not in subscription: raise AttributeError('Subscription requires a "name" key."')
        private = bool(subscription['name'] in self.private_sub_names)

        payload = {
            'event': 'unsubscribe',
            'subscription': subscription
        }
        if pair is not None:
            if not isinstance(pair, list): raise ValueError('Parameter pair must be type of List[str] (e.g. pair=["XBTUSD"])')
            payload['pair'] = pair

        if private: # private == without pair
            if not self.__is_auth: raise ValueError('Cannot unsubscribe from private feeds without valid credentials!')
            if pair is not None: raise ValueError('Cannot unsubscribe from private endpoint with specific pair!')
            await self._priv_conn.send_message(payload, private=True)

        elif pair is not None: # public with pair
            for symbol in pair:
                sub = copy.deepcopy(payload)
                sub['pair'] = [symbol]
                await self._pub_conn.send_message(sub, private=False)

        else: await self._pub_conn.send_message(payload, private=False)

    @property
    def private_sub_names(self) -> List[str]:
        '''Returns the private subscription names'''
        return  ['ownTrades', 'openOrders']

    @property
    def public_sub_names(self) -> List[str]:
        '''Returns the public subscription names'''
        return ['ticker', 'spread', 'book', 'ohlc', 'trade', '*']

    @property
    def active_public_subscriptions(self) -> List[str]:
        '''Returns the active public subscriptions'''
        if self._pub_conn is not None:
            return self._pub_conn.subscriptions
        raise ConnectionError('Public connection does not exist!')

    @property
    def active_private_subscriptions(self) -> List[str]:
        '''Returns the active private subscriptions'''
        if self._priv_conn is not None:
            return self._priv_conn.subscriptions
        raise ConnectionError('Private connection does not exist!')


